// LZSS -  
//    .
// : O(d*p)
// : O(n*p*d)
//   :
//     n -  
//     p -  
//     d -  
//  d,p ,         ,
//    .        .
//    p  log(p),       
//    ,      .
//
// icq:133575351    email:lawnmower-man@mail.ru
//
// :
// 1 N  -    
// 0 C 0 F -   ( )  8 
// 0 C 1 F -   (C )  16 

#include<stdio.h>
#include<malloc.h>
#include<memory.h>

typedef unsigned char BYTE;

//  
#define OFFSET_MIN		1		//    
#define COUNT_MIN		3		//   
#define OFFSET_BITS		16		//    
#define COUNT_BITS      7		//    
#define OFFSET1_BITS	8		//      
#define OFFSET2_BITS	16		//     

//  
#define COUNT_MAX ((1<<COUNT_BITS)-1+COUNT_MIN)		//   
#define OFFSET_MIN1 OFFSET_MIN						//   
#define OFFSET_MAX1 (OFFSET_MIN+(1<<OFFSET1_BITS)-1)//   
#define OFFSET_MIN2 OFFSET_MIN						//   
#define OFFSET_MAX2 (OFFSET_MIN+(1<<OFFSET2_BITS)-1)//   

#define BUFFER_SIZE (65536*2)		//!<    (  ,  2^n)
#if BUFFER_SIZE<(OFFSET_MAX2+1)
#  error too small buffer
#endif


class LZSS{
	int math_count;					//!<   
	int math_offset;				//!<   
	int index1,index2;				//!<    (,)
	char *buffer;					//!<  
	inline void assert(bool);
// 
	struct tree_item_st{
		int child;		//!<   
		int next;		//!<  
		int index;		//!<     
		int size;		//!<  
	}*tree;
	int tree_size;	//!<  ,   
	int tree_count;	//!<    
	int tree_root;	//!<  
	int tree_free;	//!<   
	void TreeInit();						//!<   
	int  TreeNewItem();						//!<   
	void TreeFreeItem(int);					//!<  
	void TreeRemove(int index);				//!<    
	int  TreeAdd(int index,int *offset);	//!<     
//    
	inline int GetSize();					//!<     
	inline int DiffIndex(int,int);			//!<      
	inline int IncIndex(int,int step=1);	//!<     
	inline int DecIndex(int,int step=1);	//!<     
//   
	BYTE byte,mask;
	void StartRead();						//!<   
	void StartWrite();						//!<   
	void StopWrite(FILE*);					//!<   
	int  GetBits(FILE*,int bits=1);			//!<   
	void PutBits(FILE*,int,int bits=1);		//!<  
	int GetFileSize(FILE *file);			//!<   
//    
	void FlushData(FILE*,int);				//!<     
	void FlushMath(FILE*);					//!<     
public:
	LZSS();
	~LZSS();
	void Pack(FILE*,FILE*);					//!< 
	void Depack(FILE*,FILE*);				//!< 
	static void Pack(char*,char*);
	static void Depack(char*,char*);
};

void LZSS::assert(bool x)
{
//	if(!x) _asm int 3
}


LZSS::LZSS()
{
	tree_size=64;
	buffer=(char*)malloc(BUFFER_SIZE);
	tree=(tree_item_st*)malloc(tree_size*sizeof(tree_item_st));
}

LZSS::~LZSS()
{
	free(tree);
	free(buffer);
}

//  =============================================

void LZSS::TreeInit()
{
	tree_root=tree_free=-1;
	tree_count=0;
}

//! \result   
int LZSS::TreeNewItem()
{
	int n;
	if(tree_free>=0) {
		n=tree_free;
		tree_free=tree[tree_free].next;
	} else {
		if(tree_count>=tree_size) 
			tree_size*=2,
			tree=(tree_item_st*)realloc(tree,tree_size*sizeof(tree_item_st));
		n=tree_count++;
	}
	tree[n].child=tree[n].next=-1;
	return n;
}

//! \param index   
void LZSS::TreeFreeItem(int index)
{
	tree[index].next=tree_free;
	tree_free=index;
}

//      ,    .
//       (  ,    )
//! \param index    
void LZSS::TreeRemove(int index)
{
	int i;
	int parent=-1;	//  
	int *ref=&tree_root;
	int root=tree_root;
	int count=COUNT_MAX;
	for(;;) {	//       
		for(;;) {	//   
			if(root<0) return;
			i=tree[root].index;
			if(buffer[i]==buffer[index]) break;
			ref=&tree[root].next; root=*ref;
		}
		count-=tree[root].size; if(!count) break;
		assert(count>=0);
		index=DecIndex(index,tree[root].size);
		parent=root; ref=&tree[root].child; root=*ref;
	}
	if(i!=index) return;	//       
	//   
	*ref=tree[root].next;
	TreeFreeItem(root);
	if(parent>=0) {	
		i=tree[parent].child;
		if(tree[i].next<0) {	//      ,  
			tree[parent].size+=tree[i].size;
			tree[parent].child=tree[i].child;
			TreeFreeItem(i);
		}
	}
}


//       .
// (       ,  )
//         .
//! \param index    
//! \retval offset     
//! \result  
int LZSS::TreeAdd(int index,int *offset)
{
	int n,i,nn,i0,count;
	int res=0;
	int limit=GetSize()+1; if(limit>COUNT_MAX) limit=COUNT_MAX;
	int parent=-1;
	int root=tree_root;	//   
	for(count=0;count<COUNT_MAX;) {
		i0=index;
		for(;;) {	//   
			if(root<0) {	//   
				nn=TreeNewItem();
				if(parent<0) {		//   
					tree[nn].next=tree_root;
					tree_root=nn;
				}else{				//    
					tree[nn].next=tree[parent].child;
					tree[parent].child=nn;
				}
				tree[nn].index=index;
				tree[nn].size=COUNT_MAX-count;
				return res;
			}
			i=tree[root].index;
			if(buffer[index]==buffer[i]) break;
			root=tree[root].next;
		}
		assert(i!=index);
		for(n=0; n<tree[root].size; n++) {	//   
			if(buffer[index]!=buffer[i]) { //   
				nn=TreeNewItem();		//       (       )
				tree[nn].index=i;
				tree[nn].size=tree[root].size-n;
				tree[nn].child=tree[root].child;
				tree[root].child=nn;
				tree[root].size=n;
				break;
			}
			i=DecIndex(i);
			index=DecIndex(index);
		}
		count+=n;
		assert(count<COUNT_MAX);
		if(n==tree[root].size) tree[root].index=i0;	//   ,       
		n=count;
		if(n>limit) n=limit;
		if(n<COUNT_MIN && n!=2) n=0;
		if(n>res) {
			res=n,*offset=DecIndex(index,i);
			if(res==2 && *offset>OFFSET_MAX1) res=0;
		}
		parent=root,root=tree[root].child;	//    
	}
	return res;
}

//   =============================================

int LZSS::GetFileSize(FILE *file)
{
	int size;
	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);
	return size;
}


void LZSS::StartRead()
{
	byte=mask=0;
}


void LZSS::StartWrite()
{
	byte=0,mask=1;
}

void LZSS::StopWrite(FILE* file)
{
	fputc(byte,file);
}


int  LZSS::GetBits(FILE* file,int bits)
{
	int code=0;
	for(int n=0; n<bits; n++) {
		if(!mask) byte=fgetc(file),mask=1;
		if(byte&mask) code|=1<<n;
		mask<<=1;
	}
	return code;
}


void LZSS::PutBits(FILE* file,int value,int bits)
{
	for(int n=0; n<bits; n++) {
		if(!mask) fputc(byte,file),mask=1,byte=0;
		if(value&1) byte|=mask;
		mask<<=1;
		value>>=1;
	}
}


//  =============================================

void LZSS::FlushData(FILE* file,int count)
{
	for(int n=0; n<count; n++) {
		PutBits(file,1);
		PutBits(file,buffer[index1],8);
		index1=IncIndex(index1);
	}
}


void LZSS::FlushMath(FILE *file)
{
	if(math_count<COUNT_MIN && math_count!=2) return;
	index1=IncIndex(index1,math_count);
	PutBits(file,0);
	if(math_count==2) {
		PutBits(file,1);
		PutBits(file,math_offset-OFFSET_MIN1,OFFSET1_BITS);
	}else{
		PutBits(file,0);
		PutBits(file,math_count-COUNT_MIN,COUNT_BITS);
		PutBits(file,math_offset-OFFSET_MIN2,OFFSET2_BITS);
	}
}

//   =============================================

int LZSS::DiffIndex(int i1,int i2)
{
#if BUFFER_SIZE&(BUFFER_SIZE-1)	//     
	return (i1-i2+BUFFER_SIZE)%BUFFER_SIZE;
#else
	return (unsigned int)(i1-i2)&(BUFFER_SIZE-1);
#endif
}


int LZSS::GetSize()
{
	return DiffIndex(index2,index1);
}


int LZSS::IncIndex(int index,int step)
{
#if BUFFER_SIZE&(BUFFER_SIZE-1)	//     
	return (index+step)%BUFFER_SIZE;
#else
	return (unsigned int)(index+step)&(BUFFER_SIZE-1);
#endif
}


int LZSS::DecIndex(int index,int step)
{
#if BUFFER_SIZE&(BUFFER_SIZE-1)	//     
	return (index+BUFFER_SIZE-step)%BUFFER_SIZE;
#else
	return (unsigned int)(index-step)&(BUFFER_SIZE-1);
#endif
}

//========================================================

void LZSS::Pack(FILE *input,FILE *output)
{
	int print_count=0;
	int math_count,math_offset,sym,count,n;
	index1=index2=math_count=math_offset=count=0;
	memset(buffer,0,sizeof(buffer));
	TreeInit();
	StartWrite();
	n=GetFileSize(input);
	fwrite(&n,1,sizeof(n),output);	//   
	for(;;) {
		this->math_count=math_count;
		this->math_offset=math_offset;
		sym=fgetc(input); if(sym==EOF) break;
		count++; if(count>print_count) printf("-%d   \r",count),print_count+=1243;
		TreeRemove(DecIndex(index2,OFFSET_MAX2));
		buffer[index2]=sym;
		math_count=TreeAdd(index2,&math_offset);
		index2=IncIndex(index2);
		if(this->math_count && math_count<=this->math_count) {
			FlushData(output,GetSize()-this->math_count-1);
			FlushMath(output);
			math_count=0;
		}
		n=GetSize()-math_count-COUNT_MIN;
		FlushData(output,n);
	}
	n=GetSize()-this->math_count;
	FlushData(output,n);
	FlushMath(output);
	StopWrite(output);
}


void LZSS::Depack(FILE *input,FILE *output)
{
	int sym,n,filesize,i;
	index1=index2=0;
	StartRead();
	//       
	//  ,       8    
	//     
	fread(&filesize,1,sizeof(filesize),input); 
	for(i=0;i<filesize;) {
		if(GetBits(input)) {
			sym=GetBits(input,8);
			buffer[index2]=sym; index2=IncIndex(index2); i++;
			fputc(sym,output);
		}else {
			if(GetBits(input)) math_count=2,math_offset=GetBits(input,OFFSET1_BITS)+OFFSET_MIN1;
			else 
				math_count=GetBits(input,COUNT_BITS)+COUNT_MIN,
				math_offset=GetBits(input,OFFSET2_BITS)+OFFSET_MIN2;
			for(n=0; n<math_count; n++) {
				sym=buffer[DecIndex(index2,math_offset)]; buffer[index2]=sym;
				index2=IncIndex(index2);
				fputc(sym,output);
			}
			i+=n;
		}
	}
}


void LZSS::Pack(char *inname,char *outname)
{
	LZSS obj;
	printf("Pack: %s\n",inname);
	FILE *input=fopen(inname,"rb");
	FILE *output=fopen(outname,"wb");
	if(!input || !output) {printf("IO error\n");return;}
	obj.Pack(input,output);
	fclose(input);
	fclose(output);
}


void LZSS::Depack(char *inname,char *outname)
{
	LZSS obj;
	printf("Depack: %s\n",inname);
	FILE *input=fopen(inname,"rb");
	FILE *output=fopen(outname,"wb");
	if(!input || !output) {printf("IO error\n");return;}
	obj.Depack(input,output);
	fclose(input);
	fclose(output);
}


int main(int count,char **list)
{
	if(count!=4) {
		printf("%s -e infile outfile\n",list[0]);
		printf("%s -d infile outfile\n",list[0]);
		return 1;
	}
	if(list[1][0]!='-' || list[1][1]!='e' && list[1][1]!='d') {
		printf("Syntax error\n\a");
		return 2;
	}
	switch(list[1][1]) {
	case 'e': LZSS::Pack(list[2],list[3]); break;	// 
	case 'd': LZSS::Depack(list[2],list[3]); break;	// 
	}
	return 0;
}
